import difflib
from datetime import date, datetime

import requests
from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    base_url = "https://lisburn.isl-fusion.com"

    def parse_data(self, page: str, **kwargs) -> dict:
        """
        This function will make a request to the search endpoint with the postcode, extract the
        house numbers from the responses, then retrieve the ID of the entry with the house number that matches,
        to then retrieve the bin schedule.

        The API here is a weird combination of HTML in json responses.
        """
        postcode = kwargs.get("postcode")
        paon = kwargs.get("paon")

        if not postcode:
            raise ValueError("Must provide a postcode")

        if not paon:
            raise ValueError("Must provide a house number")

        search_url = f"{self.base_url}/address/{postcode}"

        requests.packages.urllib3.disable_warnings()
        s = requests.Session()
        response = s.get(search_url)
        response.raise_for_status()

        address_data = response.json()

        address_list = address_data["html"]

        soup = BeautifulSoup(address_list, features="html.parser")

        address_by_id = {}

        for li in soup.find_all("li"):
            link = li.find_all("a")[0]
            address_id = link.attrs["href"]
            address = link.text

            address_by_id[address_id] = address

        addresses = list(address_by_id.values())

        common = difflib.SequenceMatcher(
            a=addresses[0], b=addresses[1]
        ).find_longest_match()
        extra_bit = addresses[0][common.a : common.a + common.size]

        ids_by_paon = {
            a.replace(extra_bit, ""): a_id.replace("/view/", "").replace("/", "")
            for a_id, a in address_by_id.items()
        }

        property_id = ids_by_paon.get(paon)
        if not property_id:
            raise ValueError(
                f"Invalid house number, valid values are {', '.join(ids_by_paon.keys())}"
            )

        today = date.today()
        calendar_url = (
            f"{self.base_url}/calendar/{property_id}/{today.strftime('%Y-%m-%d')}"
        )
        response = s.get(calendar_url)
        response.raise_for_status()
        calendar_data = response.json()
        next_collections = calendar_data["nextCollections"]

        collections = list(next_collections["collections"].values())

        data = {"bins": []}

        for collection in collections:
            collection_date = datetime.strptime(collection["date"], "%Y-%m-%d")
            bins = [c["name"] for c in collection["collections"].values()]

            for bin in bins:
                data["bins"].append(
                    {
                        "type": bin,
                        "collectionDate": collection_date.strftime(date_format),
                    }
                )
        return data



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get('uprn')
        check_uprn(user_uprn)
        data = {"bins": []}

        headers = {
            'authority': 'www.hull.gov.uk',
            'accept': '*/*',
            'accept-language': 'en-GB,en;q=0.9',
            'cache-control': 'no-cache',
            'pragma': 'no-cache',
            'referer': 'https://www.hull.gov.uk/bins-and-recycling/bin-collections/bin-collection-day-checker',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.6167.186 Safari/537.36',
        }
        api_url = f"https://www.hull.gov.uk/ajax/bin-collection?bindate={user_uprn}"

        res = requests.get(api_url, headers=headers)
        if res.status_code != 200:
            raise ConnectionRefusedError("Cannot connect to API!")

        json_data = res.json()[0]
        for item in json_data:
            dict_data = {
                        "type": item.get("collection_type").capitalize(),
                        "collectionDate": datetime.strptime(item.get("next_collection_date"), "%Y-%m-%d").strftime(date_format)
                    }
            data["bins"].append(dict_data)

        return data



import requests
from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        API_URLS = {
            "address_search": "https://servicelayer3c.azure-api.net/wastecalendar/address/search/",
            "collection": "https://servicelayer3c.azure-api.net/wastecalendar/collection/search/{}/",
        }
        headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.102 Safari/537.36",
        }

        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")

        # Establish session
        requests.packages.urllib3.disable_warnings()
        s = requests.Session()
        r = s.get(
            "https://www.scambs.gov.uk/recycling-and-bins/find-your-household-bin-collection-day/",
            headers=headers,
        )

        # Select address
        r = s.get(
            API_URLS["address_search"],
            headers=headers,
            params={"postCode": user_postcode},
        )
        addresses = r.json()
        address_ids = [
            x["id"] for x in addresses if x["houseNumber"].capitalize() == user_paon
        ]
        if len(address_ids) == 0:
            raise Exception(f"Could not match address {user_paon}, {user_postcode}")

        # Get the schedule
        r = s.get(
            API_URLS["collection"].format(address_ids[0]),
            headers=headers,
        )
        schedule = r.json()["collections"]

        data = {"bins": []}

        for collection in schedule:
            dt = datetime.strptime(collection["date"], "%Y-%m-%dT%H:%M:%SZ").strftime(
                date_format
            )
            for round in collection["roundTypes"]:
                dict_data = {"type": round.title(), "collectionDate": dt}
                data["bins"].append(dict_data)

        return data



import json
import logging
from jsonschema import validate, ValidationError
from pathlib import Path

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")

# Dynamically compute the base path relative to this file's location
current_file_path = Path(__file__).resolve()
BASE_PATH = current_file_path.parent.parent.parent.parent / "tests"


def load_json_file(file_name):
    file_path = BASE_PATH / file_name
    with open(file_path) as f:
        data = json.load(f)
        logging.info(f"{file_name} file loaded")
    return data


def validate_json(json_str):
    try:
        return json.loads(json_str)
    except ValueError as err:
        logging.error(f"JSON validation error: {err}")
        raise


def validate_json_schema(json_str, schema):
    json_data = validate_json(json_str)
    try:
        validate(instance=json_data, schema=schema)
    except ValidationError as err:
        logging.error(f"Schema validation error: {err}")
        logging.info(f"Data: {json_str}")
        logging.info(f"Schema: {schema}")
        raise
    return True



import logging
import pytest
import traceback
from pytest_bdd import scenario, given, when, then, parsers
from hamcrest import assert_that, equal_to
from functools import wraps

from step_helpers import file_handler
from uk_bin_collection.uk_bin_collection import collect_data

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


@scenario("../features/validate_council_outputs.feature", "Validate Council Output")
def test_scenario_outline():
    pass


def handle_test_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.error(f"Error in test '{func.__name__}': {e}")
            logging.error(traceback.format_exc())
            raise e

    return wrapper


@pytest.fixture
def context():
    class Context(object):
        pass

    return Context()


@handle_test_errors
@given(parsers.parse("the council: {council_name}"))
def get_council_step(context, council_name):
    council_input_data = file_handler.load_json_file("input.json")
    context.metadata = council_input_data[council_name]


# When we scrape the data from <council> using <selenium_mode> and the <selenium_url> is set.
@handle_test_errors
@when(
    parsers.parse(
        "we scrape the data from {council} using {selenium_mode} and the {selenium_url} is set"
    )
)
def scrape_step(context, council, selenium_mode, selenium_url):
    context.council = council
    context.selenium_mode = selenium_mode
    context.selenium_url = selenium_url

    args = [council, context.metadata["url"]]

    if "uprn" in context.metadata:
        uprn = context.metadata["uprn"]
        args.append(f"-u={uprn}")
    if "postcode" in context.metadata:
        postcode = context.metadata["postcode"]
        args.append(f"-p={postcode}")
    if "house_number" in context.metadata:
        house_number = context.metadata["house_number"]
        args.append(f"-n={house_number}")
    if "usrn" in context.metadata:
        usrn = context.metadata["usrn"]
        args.append(f"-us={usrn}")
    if "headless" in context.metadata:
        headless = context.metadata["headless"]
        if headless:
            args.append(f"--headless")
        else:
            args.append(f"--not-headless")
    # TODO we should somehow run this test with and without this argument passed
    # TODO I do think this would make the testing of the councils a lot longer and cause a double hit from us

    # At the moment the feature file is set to local execution of the selenium so no url will be set
    # And it the behave test will execute locally
    if selenium_mode != "None" and selenium_url != "None":
        if selenium_mode != "local":
            args.append(f"-w={selenium_url}")
    if "skip_get_url" in context.metadata:
        args.append(f"-s")

    CollectData = collect_data.UKBinCollectionApp()
    CollectData.set_args(args)
    context.parse_result = CollectData.run()


@handle_test_errors
@then("the result is valid json")
def validate_json_step(context):
    assert file_handler.validate_json(context.parse_result), "Invalid JSON output"


@handle_test_errors
@then("the output should validate against the schema")
def validate_output_step(context):
    council_schema = file_handler.load_json_file(f"output.schema")
    assert file_handler.validate_json_schema(
        context.parse_result, council_schema
    ), "Schema validation failed"



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Make a BS4 object
        soup = BeautifulSoup(page.text, features="html.parser")
        soup.prettify()

        data = {"bins": []}
        collections = []

        bins = soup.find("div", {"class": "results-table-wrapper"}).find_all("div", {"class": "service-wrapper"})
        for bin in bins:
            bin_type = bin.find("h3", {"class": "service-name"}).get_text().replace("Collection", "bin").strip()
            bin_date = datetime.strptime(bin.find("td", {"class": "next-service"}).find("span", {"class": "table-label"}).next_sibling.get_text().strip(), "%d/%m/%Y")
            collections.append((bin_type, bin_date))

        ordered_data = sorted(collections, key=lambda x: x[1])
        for item in ordered_data:
            dict_data = {
                "type": item[0].capitalize(),
                "collectionDate": item[1].strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data



import urllib.parse

from bs4 import BeautifulSoup
from dateutil.relativedelta import relativedelta

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        data = {"bins": []}

        user_postcode = kwargs.get("postcode")
        user_number = kwargs.get("paon")

        user_address = "{} {}".format(user_number, user_postcode)
        user_address = urllib.parse.quote(user_address)

        # This first URL checks against a string representing the users address and returns values used for a second lookup.
        stage1_url = "https://wlnk.statmap.co.uk/map/Cluster.svc/findLocation?callback=getAddressesCallback1702938375023&script=%5CCluster%5CCluster.AuroraScript%24&address={}".format(
            user_address
        )

        address_data = requests.get(stage1_url).text

        # Strip data and parse the JSON
        address_data = json.loads(
            re.sub("getAddressesCallback\d+\(", "", address_data)[:-2]
        )

        if address_data["TotalHits"] == 0:
            raise Exception(
                "No address found for string {}. See Wiki".format(user_address)
            )
        elif address_data["TotalHits"] != 1:
            # Multiple hits returned. Let's pick the first one. We could raise an exception here if this causes problems.
            pass

        # Pull out the address data needed for the next step
        address_id = address_data["Locations"][0]["Id"]
        address_x = address_data["Locations"][0]["X"]
        address_y = address_data["Locations"][0]["Y"]

        stage2_url = "https://wlnk.statmap.co.uk/map/Cluster.svc/getpage?script=\Cluster\Cluster.AuroraScript$&taskId=bins&format=js&updateOnly=true&query=x%3D{}%3By%3D{}%3Bid%3D{}".format(
            address_x, address_y, address_id
        )

        bin_query = requests.get(stage2_url).text

        # Test that what we got is good
        if "injectCss" not in bin_query:
            raise Exception(
                "Error. Data has not been returned correctly. Please raise an issue on the GitHub page"
            )

        # Return only the HTML contained within the Javascript function payload.
        pattern = 'document\.getElementById\("DR1"\)\.innerHTML="(.+)";'

        bin_html = re.findall(pattern, bin_query)

        if len(bin_html) != 1:
            # This exception is raised if the regular expression above finds anything other than one expected match.
            raise Exception(
                "Incorrect number of matches found during phase 2 search. Please raise an issue on the Github page"
            )

        # Some silly python foo required here to unescape the unicode contained.
        bin_html = bin_html[0].encode().decode("unicode-escape")

        soup = BeautifulSoup(bin_html, "html.parser")

        collection_rows = soup.find("li", {"class": "auroraListItem"}).find_all("li")

        collections = []

        for row in collection_rows:
            # Get bin type
            bin_type = row.find("span").text

            # Get bin date
            bin_date_text = row.text
            pattern = "\d+\/\d+"
            bin_dates = re.findall(pattern, bin_date_text)

            for bin_date in bin_dates:
                # Split the bin date into day and month and build a full date with the current year
                split_date = bin_date.split('/')
                full_date = datetime(datetime.now().year, int(split_date[1]), int(split_date[0]))

                # If the current month is December and one of the next collections is in January, increment the year
                if datetime.now().month == 12 and int(split_date[1]) == 1:
                    full_date = bin_date + relativedelta(years=1)

                # Since data in unordered, add to a tuple
                collections.append((bin_type.title(), full_date))

        # Sort the tuple by date
        ordered_data = sorted(collections, key=lambda x: x[1])

        # Add everything into the dictionary
        for item in ordered_data:
            dict_data = {
                "type": item[0],
                "collectionDate": item[1].strftime(date_format),
            }

            data["bins"].append(dict_data)

        return data



from bs4 import BeautifulSoup
from datetime import datetime
import re
import requests
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass

# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_postcode = kwargs.get("postcode")
        check_postcode(user_postcode)

        root_url = "https://molevalley.cloudmappin.com/my-mv-address-search/search/{}/0".format(user_postcode)
        response = requests.get(root_url)

        if not response.ok:
            raise ValueError("Invalid server response code retreiving data.")

        jsonData = response.json()

        if len(jsonData["results"]) == 0:
            raise ValueError("No collection data found for postcode provided.")

        properties_found = jsonData["results"][0]["items"]

        # If UPRN is provided, we can check a specific address.
        html_data = None
        uprn = kwargs.get("uprn")
        if uprn:
            check_uprn(uprn)
            for n, item in enumerate(properties_found):
                if uprn == str(int(item["info"][0][1]["value"])):
                    html_data = properties_found[n]["info"][2][1]["value"]
                    break
            if html_data is None:
                raise ValueError("No collection data found for UPRN provided.")
        else:
            # If UPRN not provided, just use the first result
            html_data = properties_found[0]["info"][2][1]["value"]

        soup = BeautifulSoup(html_data, features="html.parser")
        soup.prettify()

        data = {"bins": []}
        all_collection_dates = []
        regex_date = re.compile(r'.* ([\d]+\/[\d]+\/[\d]+)')
        regex_additional_collection = re.compile(r'We also collect (.*) on (.*) -')

        # Search for the 'Bins and Recycling' panel
        for panel in soup.select('div[class*="panel"]'):
            if panel.h2.text.strip() == "Bins and Recycling":

                # Gather the bin types and dates
                for collection in panel.select('div > strong'):
                    bin_type = collection.text.strip()
                    collection_string = collection.find_next('p').text.strip()
                    m = regex_date.match(collection_string)
                    if m:
                        collection_date = datetime.strptime(m.group(1),'%d/%m/%Y').date()
                        data["bins"].append({"type": bin_type, "collectionDate": collection_date.strftime('%d/%m/%Y')})
                        all_collection_dates.append(collection_date)

                # Search for additional collections
                for p in panel.select('p'):
                    m2 = regex_additional_collection.match(p.text.strip())
                    if m2:
                        bin_type = m2.group(1)
                        if "each collection day" in m2.group(2):
                            collection_date = min(all_collection_dates)
                            data["bins"].append({"type": bin_type, "collectionDate": collection_date.strftime('%d/%m/%Y')})
                break

        return data
